unit pph;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics,
  pphbits, pphtypes;

const
  PPH_max_parts = 12*12;
type
  TPPH = class (TObject)
  private
//public
    FSourceSize: TPoint;
    FSize: Integer;
    FY,FCb,FCr: PIntegerArray;
    // parts
    FPartsNum: Integer;
    FPartPos, FPartSize: array [0..PPH_max_parts-1] of TPoint;
    FPartCount: array [0..PPH_max_parts-1] of Integer;
    // bits
    FS_Q: Integer;
    Fbits: TPPHbits;
    //
    procedure FCalcParts;
    procedure FSetSourceSize (new_size: TPoint);
    procedure FMountPart (source: PIntegerArrays; pos,size: TPoint; dest: PIntegerArray);
    procedure FUnmountPart (dest: PIntegerArrays; pos,size: TPoint; source: PIntegerArray);
    procedure FCreamizePart (values: PIntegerArray; pos,size: Integer);
    procedure FUncreamizePart (values: PIntegerArray; pos,size: Integer);
    {function FMistake (a1, a2: PIntegerArray; size: Integer): Integer;}
    //
    procedure FPutCreamOfFirst (cream,dest: PIntegerArray; size, first: Integer);
    procedure FGetCreamOfFirst (dest: PIntegerArray; size: Integer);
    {function FFindFirstOfSize (values: PIntegerArray; size_limit: Integer): Integer;}
    {function FMistakeOfFirst (cream,original: PIntegerArray; size, first: Integer): Integer;
    function FFindFirstOfMistake (values: PIntegerArray; size, mistake: Integer): Integer;}
    function FQPartOfFirst (values: PIntegerArray; size, first: Integer): Integer;
    function FFindFirstOfQPart (values: PIntegerArray; size, qpart: Integer): Integer;
  public
    {}sizeY,sizeCb,sizeCr: Integer;
    {}numqY,numqCb,numqCr: Integer;
    constructor Create;
    destructor Destroy; override;
    procedure FromBitmap (bmp: TBitmap);
    procedure ToBitmap (bmp: TBitmap);
    procedure CreamizeAll;
    procedure UncreamizeAll;
    {function SaveCreamOfSize (size_limit: Integer; dest: TPPH): PIntegerArray;}
    {function SaveCreamOfMistake (dest: TPPH; mtkY,mtkCb,mtkCr: Integer): PIntegerArray;}
    procedure SaveCreamOfQPart (port: TPPHbits; dest: TPPH; qY,qCb,qCr: Integer);
    procedure LoadCream (port: TPPHbits);
    procedure LoadFromFile (filename: string);
    property SourceSize: TPoint read FSourceSize write FSetSourceSize;
    property lY: PIntegerArray read FY;
    property lCb: PIntegerArray read FCb;
    property lCr: PIntegerArray read FCr;
    property NumPixels: Integer read FSize;
    property bits_port: TPPHbits write Fbits;
  end;

implementation

constructor TPPH.Create;
begin
  inherited Create;
  FY := nil;
  FCb := nil;
  FCr := nil;
end;

destructor TPPH.Destroy;
begin
  if FY <> nil then
    FreeMem (FY);
  if FCb <> nil then
    FreeMem (FCb);
  if FCr <> nil then
    FreeMem (FCr);
  inherited Destroy;
end;

procedure TPPH.FCalcParts;
var
  pos, size: TPoint;
  tsizex: Integer;
begin
  FPartsNum := 0;
  //
  size := Point(1,1);
  while (size.x*2) <= SourceSize.x do
    size.x := size.x*2;
  while (size.y*2) <= SourceSize.y do
    size.y := size.y*2;
  tsizex := size.x;
  pos.y := 0;
  //
  while size.y <> 0 do begin
    if (size.y and SourceSize.y) <> 0 then begin
      pos.x := 0;
      size.x := tsizex;
      while size.x <> 0 do begin
        if (size.x and SourceSize.x) <> 0 then begin
          FPartPos[FPartsNum] := pos;
          FPartSize[FPartsNum] := size;
          FPartCount[FPartsNum] := size.x*size.y;
          Inc (FPartsNum);
          pos.x := pos.x + size.x;
        end;
        size.x := size.x div 2;
      end;
      pos.y := pos.y + size.y;
    end;
    size.y := size.y div 2;
  end;
end;

procedure TPPH.FSetSourceSize (new_size: TPoint);
begin
  FSize := new_size.x*new_size.y;
  ReAllocMem (FY, FSize*SizeOf(Integer));
  ReAllocMem (FCb, FSize*SizeOf(Integer));
  ReAllocMem (FCr, FSize*SizeOf(Integer));
  FSourceSize := new_size;
  FCalcParts;
end;

procedure TPPH.FMountPart (source: PIntegerArrays; pos,size: TPoint; dest: PIntegerArray);
begin
  if (size.x = 1) and (size.y = 1) then
    dest[0] := source[pos.y][pos.x]
  else if size.x > size.y then begin
    size.x := size.x div 2;
    FMountPart (source, pos, size, dest);
    pos.x := pos.x + size.x;
    Integer(dest) := Integer(dest) + size.x*size.y*SizeOf(Integer);
    FMountPart (source, pos, size, dest);
  end else begin
    size.y := size.y div 2;
    FMountPart (source, pos, size, dest);
    pos.y := pos.y + size.y;
    Integer(dest) := Integer(dest) + size.x*size.y*SizeOf(Integer);
    FMountPart (source, pos, size, dest);
  end;
end;

procedure TPPH.FUnmountPart (dest: PIntegerArrays; pos,size: TPoint; source: PIntegerArray);
begin
  if (size.x = 1) and (size.y = 1) then
    dest[pos.y][pos.x] := source[0]
  else if size.x > size.y then begin
    size.x := size.x div 2;
    FUnMountPart (dest, pos, size, source);
    pos.x := pos.x + size.x;
    Integer(source) := Integer(source) + size.x*size.y*SizeOf(Integer);
    FUnMountPart (dest, pos, size, source);
  end else begin
    size.y := size.y div 2;
    FUnMountPart (dest, pos, size, source);
    pos.y := pos.y + size.y;
    Integer(source) := Integer(source) + size.x*size.y*SizeOf(Integer);
    FUnMountPart (dest, pos, size, source);
  end;
end;

procedure TPPH.FCreamizePart (values: PIntegerArray; pos,size: Integer);
var
  size2, n: Integer;
  sum,dif: PIntegerArray;
begin
  if size > 2 then begin
    // do this
    size2 := size div 2;
    sum := AllocIntArray (size2);
    dif := AllocIntArray (size2);
    while size2 > 0 do begin
      for n := size2-1 downto 0 do begin
        sum[n] := values[pos+n*2+1] + values[pos+n*2+0];
        dif[n] := values[pos+n*2+1] - values[pos+n*2+0];
      end;
      for n := size2-1 downto 0 do begin
        values[pos+n] := sum[n];
        values[pos+size2+n] := dif[n];
      end;
      // recurcy forever
      if size2 >= 2 then
        FCreamizePart (values, pos+size2, size2);
      size2 := size2 div 2;
    end;
    FreeMem (dif);
    FreeMem (sum);
  end else if size = 2 then begin
    Inc (values[pos+0], values[pos+1]);
    values[pos+1] := values[pos+1]*2 - values[pos+0];
  end;
end;

procedure TPPH.FUncreamizePart (values: PIntegerArray; pos,size: Integer);
var
  size2, n: Integer;
  sum,dif: PIntegerArray;
begin
  if size > 2 then begin
    // do this
    size2 := 1;
    sum := AllocIntArray (size div 2);
    dif := AllocIntArray (size div 2);
    while size2 < size do begin
      // recurcy forever
      if size2 >= 2 then
        FUncreamizePart (values, pos+size2, size2);
      //
      for n := 0 to size2-1 do begin
        sum[n] := values[pos+n];
        dif[n] := values[pos+size2+n];
      end;
      for n := 0 to size2-1 do begin
        values[pos+n*2+1] := (sum[n] + dif[n]) div 2;
        values[pos+n*2+0] := sum[n] - values[pos+n*2+1];
      end;
      size2 := size2*2;
    end;
    FreeMem (dif);
    FreeMem (sum);
  end else if size = 2 then begin
    values[pos+1] := (values[pos+0] + values[pos+1]) div 2;
    values[pos+0] := values[pos+0] - values[pos+1];
  end;
end;
{
function TPPH.FMistake (a1, a2: PIntegerArray; size: Integer): Integer;
var
  n, d: Integer;
begin
  Result := 0;
  for n := 0 to size-1 do begin
    d := a1[n] - a2[n];
    d := d*d;
    //if d > Result then
      //Result := d;
    Inc (Result, d*d);
  end;
  Result := Result div size;
end;
}
procedure TPPH.FromBitmap (bmp: TBitmap);
var
  bmp_line: PRGBArray;
  bmpY,bmpCb,bmpCr: PIntegerArrays;
  c: TYCbCr;
  n, x,y: Integer;
  aY,aCb,aCr: PIntegerArray;
begin
  SourceSize := Point (bmp.Width, bmp.Height);
  // transform to YCbCr - format
  bmpY := AllocIntArrays (bmp.Width, bmp.Height);
  bmpCb := AllocIntArrays (bmp.Width, bmp.Height);
  bmpCr := AllocIntArrays (bmp.Width, bmp.Height);
  bmp.PixelFormat := pf24bit;
  for y := 0 to bmp.Height-1 do begin
    bmp_line := bmp.ScanLine[y];
    for x := 0 to bmp.Width-1 do begin
      c := RGB2YCbCr (bmp_line[x]);
      bmpY[y][x] := c.Y;
      bmpCb[y][x] := c.Cb;
      bmpCr[y][x] := c.Cr;
    end;
  end;
  // and other trans..
  aY := FY;
  aCb := FCb;
  aCr := FCr;
  //
  for n := 0 to FPartsNum-1 do begin
    FMountPart (bmpY,  FPartPos[n], FPartSize[n], aY);
    FMountPart (bmpCb, FPartPos[n], FPartSize[n], aCb);
    FMountPart (bmpCr, FPartPos[n], FPartSize[n], aCr);
    {q := trunc(sqrt (max_square div q));
    for n := 0 to size.x*size.y-1 do begin
      aY[n] := aY[n]*q;
      aCb[n] := aCb[n]*q;
      aCr[n] := aCr[n]*q;
    end;}
    Integer(aY)  := Integer(aY)  + FPartCount[n]*SizeOf(Integer);
    Integer(aCb) := Integer(aCb) + FPartCount[n]*SizeOf(Integer);
    Integer(aCr) := Integer(aCr) + FPartCount[n]*SizeOf(Integer);
  end;
  // freemem
  FreeMem (bmpY);
  FreeMem (bmpCb);
  FreeMem (bmpCr);
end;

procedure TPPH.ToBitmap (bmp: TBitmap);
var
  bmp_line: PRGBArray;
  bmpY,bmpCb,bmpCr: PIntegerArrays;
  c: TYCbCr;
  n, x,y: Integer;
  aY,aCb,aCr: PIntegerArray;
begin
  bmpY := AllocIntArrays (SourceSize.x, SourceSize.y);
  bmpCb := AllocIntArrays (SourceSize.x, SourceSize.y);
  bmpCr := AllocIntArrays (SourceSize.x, SourceSize.y);
  // transform
  aY := FY;
  aCb := FCb;
  aCr := FCr;
  //
  for n := 0 to FPartsNum-1 do begin
    FUnmountPart (bmpY,  FPartPos[n], FPartSize[n], aY);
    FUnmountPart (bmpCb, FPartPos[n], FPartSize[n], aCb);
    FUnmountPart (bmpCr, FPartPos[n], FPartSize[n], aCr);
    {q := trunc(sqrt (max_square div q));
    for y := pos.y to pos.y+size.y-1 do
      for x := pos.x to pos.x+size.x-1 do begin
        vY[y][x] := vY[y][x] div q;
        vCb[y][x] := vCb[y][x] div q;
        vCr[y][x] := vCr[y][x] div q;
      end;}
    Integer(aY)  := Integer(aY)  + FPartCount[n]*SizeOf(Integer);
    Integer(aCb) := Integer(aCb) + FPartCount[n]*SizeOf(Integer);
    Integer(aCr) := Integer(aCr) + FPartCount[n]*SizeOf(Integer);
  end;
  // transform to RGB - format
  bmp.Width := SourceSize.x;
  bmp.Height := SourceSize.y;
  bmp.PixelFormat := pf24bit;
  for y := 0 to bmp.Height-1 do begin
    bmp_line := bmp.ScanLine[y];
    for x := 0 to bmp.Width-1 do begin
      c.Y := bmpY[y][x];
      c.Cb := bmpCb[y][x];
      c.Cr := bmpCr[y][x];
      bmp_line[x] := YCbCr2RGB (c);
    end;
  end;
  // freemem
  FreeMem (bmpY);
  FreeMem (bmpCb);
  FreeMem (bmpCr);
end;

procedure TPPH.CreamizeAll;
var
  aY,aCb,aCr: PIntegerArray;
  n, s: Integer;
begin
  aY := FY;
  aCb := FCb;
  aCr := FCr;
  //
  for n := 0 to FPartsNum-1 do begin
    s := FPartCount[n];
    FCreamizePart (aY,  0, s);
    FCreamizePart (aCb, 0, s);
    FCreamizePart (aCr, 0, s);
    s := s*SizeOf(Integer);
    Integer(aY)  := Integer(aY)  + s;
    Integer(aCb) := Integer(aCb) + s;
    Integer(aCr) := Integer(aCr) + s;
  end;
end;

procedure TPPH.UncreamizeAll;
var
  aY,aCb,aCr: PIntegerArray;
  n, s: Integer;
begin
  aY := FY;
  aCb := FCb;
  aCr := FCr;
  //
  for n := 0 to FPartsNum-1 do begin
    s := FPartCount[n];
    FUncreamizePart (aY,  0, s);
    FUncreamizePart (aCb, 0, s);
    FUncreamizePart (aCr, 0, s);
    s := s*SizeOf(Integer);
    Integer(aY)  := Integer(aY)  + s;
    Integer(aCb) := Integer(aCb) + s;
    Integer(aCr) := Integer(aCr) + s;
  end;
end;

//

procedure TPPH.FPutCreamOfFirst (cream,dest: PIntegerArray; size, first: Integer);
var
  first2: Integer;
  bits: PIntegerArray;
  n, znum, znum_last: Integer;
  level, level_last: Integer;
  bits_pos: Integer;
begin
  first2 := first div 2;
  Fbits.PutAbsValue (first, size*255);
  //
  for n := 0 to size-1 do
    dest[n] := 0;
  //
  bits := AllocIntArray ((size+31) div 32);
  // levels
  level := first;
  level_last := 0;
  znum_last := size;
  while znum_last <> 0 do begin
    for n := 0 to ((znum_last+31) div 32)-1 do
      bits[n] := 0;
    bits_pos := 0;
    znum := 0;
    for n := 0 to size-1 do
      if dest[n] = level_last then begin
        if (first2 + Abs(cream[n])) >= level then begin
          dest[n] := level;
          Inc (bits[bits_pos shr 5], 1 shl (bits_pos and $1f));
          Inc (znum);
        end;
        Inc (bits_pos);
      end;
    Fbits.CompressBits (bits, znum_last, znum);
    znum_last := znum;
    level_last := level;
    if level = first then
      Inc (FS_Q, znum);
    level := level*2;
  end;
  // cream
  for n := 0 to size-1 do
    if dest[n] <> 0 then begin
      level := dest[n];
      while level > first do begin
        level := level div 2;
        if (dest[n] + level) <= (first2 + Abs(cream[n])) then begin
          Inc (dest[n], level);
          Fbits.PutBit (True);
        end else
          Fbits.PutBit (False);
      end;
    end;
  // +/-
  for n := 0 to size-1 do
    if dest[n] <> 0 then
      if cream[n] > 0 then begin
        Fbits.PutBit (True);
      end else begin
        Fbits.PutBit (False);
        dest[n] := -dest[n];
      end;
  //
  FreeMem (bits);
end;

procedure TPPH.FGetCreamOfFirst (dest: PIntegerArray; size: Integer);
var
  first: Integer;
  bits: PIntegerArray;
  n, znum, znum_last: Integer;
  level, level_last: Integer;
  bits_pos: Integer;
begin
  // get first
  first := Fbits.GetAbsValue (size*255);
  //
  for n := 0 to size-1 do
    dest[n] := 0;
  // levels
  level := first;
  level_last := 0;
  znum_last := size;
  while znum_last <> 0 do begin
    bits := Fbits.DecompressBits (znum_last);
    bits_pos := 0;
    znum := 0;
    for n := 0 to size-1 do
      if dest[n] = level_last then begin
        if (bits[bits_pos shr 5] and (1 shl (bits_pos and $1f))) <> 0 then begin
          dest[n] := level;
          Inc (znum);
        end;
        Inc (bits_pos);
      end;
    FreeMem (bits);
    znum_last := znum;
    level_last := level;
    if level = first then
      Inc (FS_Q, znum);
    level := level*2;
  end;
  // cream
  for n := 0 to size-1 do
    if dest[n] <> 0 then begin
      level := dest[n];
      while level > first do begin
        level := level div 2;
        if Fbits.GetBit then
          Inc (dest[n], level);
      end;
    end;
  // +/-
  for n := 0 to size-1 do
    if dest[n] <> 0 then
      if not Fbits.GetBit then
        dest[n] := -dest[n];
end;
{
function TPPH.FFindFirstOfSize (values: PIntegerArray; size_limit: Integer): Integer;
var
  dest: PIntegerArray;
  v1, v1_size: Integer;
  v2, v2_size, v: Integer;
begin
  dest := AllocIntArray (FSize);
  v1 := 1;
  v2 := 2;
  //
  FPutBitInit;
  v2_size := FPutCream (values, dest, FSize, v2);
  while v2_size > size_limit do begin
    v1 := v2;
    //v1_size := v2_size;
    v2 := v2*2;
    FPutBitInit;
    v2_size := FPutCreamOfFirst (values, dest, FSize, v2);
  end;
  //
  while (v2-v1) > 1 do begin
    v := (v2+v1) div 2;
    FPutBitInit;
    if FPutCreamOfFirst (values, dest, FSize, v) > size_limit then
      v1 := v
    else
      v2 := v;
  end;
  //
  FreeMem (dest);
  FreeMem (Fbits_array);
  //
  Result := v2;
end;
}{
function TPPH.FMistakeOfFirst (cream,original: PIntegerArray; size, first: Integer): Integer;
var
  cream2: PIntegerArray;
  bits: TPPHbits;
begin
  bits := TPPHbits.Create;
  cream2 := AllocIntArray (size);
  // calc puted size
  bits.PutBitInit;
  FPutCreamOfFirst (cream, cream2, size, first);
  //FreeMem (Fbits_array);
  // uncream result
  FUncreamizePart (cream2, 0, size);
  Result := FMistake (cream2, original, size);
  // free mem
  FreeMem (cream2);
end;
}{
function TPPH.FFindFirstOfMistake (values: PIntegerArray; size, mistake: Integer): Integer;
var
  cream: PIntegerArray;
  n: Integer;
  f1,f,f2: Integer;
begin
  // make cream
  cream := AllocIntArray (size);
  for n := 0 to size-1 do
    cream[n] := values[n];
  FCreamizePart (cream, 0, size);
  // lesson 1
  f1 := 0;
  f2 := 1;
  while (FMistakeOfFirst (cream, values, size, f2) <= mistake) and (f2 < $ffffff) do begin
    f1 := f2;
    f2 := f1*2;
  end;
  // lesson 2
  while (f2 - f1) > 1 do begin
    f := (f1 + f2) div 2;
    if FMistakeOfFirst (cream, values, size, f) <= mistake then
      f1 := f
    else
      f2 := f;
  end;
  Result := f2;
  // free mem
  FreeMem (cream);
end;
}
function TPPH.FQPartOfFirst (values: PIntegerArray; size, first: Integer): Integer;
var
  nz, n: Integer;
  first2: Integer;
begin
  first2 := first - (first div 2);
  nz := 0;
  Result := 0;
  for n := 0 to size-1 do
    if values[n] <> 0 then begin
      Inc (nz);
      if Abs(values[n]) >= first2 then
        Inc (Result);
    end;
  if nz <> 0 then
    if Result < (nz - Result) then
      Result := 1000*Result div nz
    else
      Result := 1000 - (1000*(nz - Result) div nz)
  else
    Result := 1000;
end;

function TPPH.FFindFirstOfQPart (values: PIntegerArray; size, qpart: Integer): Integer;
var
  f1,f,f2: Integer;
begin
  // lesson 1
  f1 := 0;
  f2 := 1;
  while (FQPartOfFirst (values, size, f2) >= qpart) and (f2 < $ffffff) do begin
    f1 := f2;
    f2 := f1*2;
  end;
  // lesson 2
  while (f2 - f1) > 1 do begin
    f := (f1 + f2) div 2;
    if FQPartOfFirst (values, size, f) >= qpart then
      f1 := f
    else
      f2 := f;
  end;
  Result := f2;
end;
{
function TPPH.SaveCreamOfMistake (dest: TPPH; mtkY,mtkCb,mtkCr: Integer): PIntegerArray;
var
  sY,sCb,sCr: PIntegerArray;
  dY,dCb,dCr: PIntegerArray;
  firstY,firstCb,firstCr: array [0..PPH_max_parts-1] of Integer;
  n, size: Integer;
  //bpos,qpos: Integer;
begin
  // set source size of dest
  dest.SourceSize := SourceSize;
  // calc first-values of parts
  sY  := FY;
  sCb := FCb;
  sCr := FCr;
  for n := 0 to FPartsNum-1 do begin
    size := FPartCount[n];
    firstY[n]  := FFindFirstOfMistake (sY,  size, mtkY);
    firstCb[n] := FFindFirstOfMistake (sCb, size, mtkCb);
    firstCr[n] := FFindFirstOfMistake (sCr, size, mtkCr);
    size := size*SizeOf(Integer);
    Inc (Integer(sY),  size);
    Inc (Integer(sCb), size);
    Inc (Integer(sCr), size);
  end;
  // save cream of parts
  CreamizeAll;
  FPutBitInit;
  // put size
  FPutAbsValue (SourceSize.x, $fff);
  FPutAbsValue (SourceSize.y, $fff);
  //sizeY  := 0;
  //sizeCb := 0;
  //sizeCr := 0;
  //bpos := 0;
  //numqY  := 0;
  //numqCb := 0;
  //numqCr := 0;
  //FS_Q := 0;
  //qpos := FS_Q;
  sY  := FY;  dY  := dest.lY;
  sCb := FCb; dCb := dest.lCb;
  sCr := FCr; dCr := dest.lCr;
  for n := 0 to FPartsNum-1 do begin
    size := FPartCount[n];
    FPutCreamOfFirst (sY,  dY,  size, firstY[n]);
    //Inc (sizeY, Fbits_pos - bpos);
    //bpos := Fbits_pos;
    //Inc (numqY, FS_Q - qpos);
    //qpos := FS_Q;
    FPutCreamOfFirst (sCb, dCb, size, firstCb[n]);
    //Inc (sizeCb, Fbits_pos - bpos);
    //bpos := Fbits_pos;
    //Inc (numqCb, FS_Q - qpos);
    //qpos := FS_Q;
    FPutCreamOfFirst (sCr, dCr, size, firstCr[n]);
    //Inc (sizeCr, Fbits_pos - bpos);
    //bpos := Fbits_pos;
    //Inc (numqCr, FS_Q - qpos);
    //qpos := FS_Q;
    size := size*SizeOf(Integer);
    Inc (Integer(sY),  size); Inc (Integer(dY),  size);
    Inc (Integer(sCb), size); Inc (Integer(dCb), size);
    Inc (Integer(sCr), size); Inc (Integer(dCr), size);
  end;
  // uncream source and destination
  UncreamizeAll;
  dest.UncreamizeAll;
  // send result
  Result := Fbits_array;
end;
}
procedure TPPH.SaveCreamOfQPart (port: TPPHbits; dest: TPPH; qY,qCb,qCr: Integer);
var
  sY,sCb,sCr: PIntegerArray;
  dY,dCb,dCr: PIntegerArray;
  firstY,firstCb,firstCr: array [0..PPH_max_parts-1] of Integer;
  n, size: Integer;
  {}bpos,qpos: Integer;
begin
  // set source size of dest
  dest.SourceSize := SourceSize;
  // cream this
  CreamizeAll;
  // calc first-values of parts
  sY  := FY;
  sCb := FCb;
  sCr := FCr;
  for n := 0 to FPartsNum-1 do begin
    size := FPartCount[n];
    firstY[n]  := FFindFirstOfQPart (sY,  size, qY);
    firstCb[n] := FFindFirstOfQPart (sCb, size, qCb);
    firstCr[n] := FFindFirstOfQPart (sCr, size, qCr);
    size := size*SizeOf(Integer);
    Inc (Integer(sY),  size);
    Inc (Integer(sCb), size);
    Inc (Integer(sCr), size);
  end;
  // save cream of parts
  Fbits := port;
  Fbits.PutBitInit;
  // put size
  Fbits.PutAbsValue (SourceSize.x, $fff);
  Fbits.PutAbsValue (SourceSize.y, $fff);
  {}sizeY  := 0;
  {}sizeCb := 0;
  {}sizeCr := 0;
  {}bpos := 0;
  {}numqY  := 0;
  {}numqCb := 0;
  {}numqCr := 0;
  {}FS_Q := 0;
  {}qpos := FS_Q;
  sY  := FY;  dY  := dest.lY;
  sCb := FCb; dCb := dest.lCb;
  sCr := FCr; dCr := dest.lCr;
  for n := 0 to FPartsNum-1 do begin
    size := FPartCount[n];
    FPutCreamOfFirst (sY,  dY,  size, firstY[n]);
    {}Inc (sizeY, Fbits.bits_pos - bpos);
    {}bpos := Fbits.bits_pos;
    {}Inc (numqY, FS_Q - qpos);
    {}qpos := FS_Q;
    FPutCreamOfFirst (sCb, dCb, size, firstCb[n]);
    {}Inc (sizeCb, Fbits.bits_pos - bpos);
    {}bpos := Fbits.bits_pos;
    {}Inc (numqCb, FS_Q - qpos);
    {}qpos := FS_Q;
    FPutCreamOfFirst (sCr, dCr, size, firstCr[n]);
    {}Inc (sizeCr, Fbits.bits_pos - bpos);
    {}bpos := Fbits.bits_pos;
    {}Inc (numqCr, FS_Q - qpos);
    {}qpos := FS_Q;
    size := size*SizeOf(Integer);
    Inc (Integer(sY),  size); Inc (Integer(dY),  size);
    Inc (Integer(sCb), size); Inc (Integer(dCb), size);
    Inc (Integer(sCr), size); Inc (Integer(dCr), size);
  end;
  // uncream source and destination
  UncreamizeAll;
  dest.UncreamizeAll;
end;

procedure TPPH.LoadCream (port: TPPHbits);
var
  aY,aCb,aCr: PIntegerArray;
  n, size: Integer;
  ssize: TPoint;
begin
  Fbits := port;
  // get size
  ssize.x := Fbits.GetAbsValue($fff);
  ssize.y := Fbits.GetAbsValue($fff);
  SourceSize := ssize;
  //
  aY  := FY;
  aCb := FCb;
  aCr := FCr;
  for n := 0 to FPartsNum-1 do begin
    size := FPartCount[n];
    FGetCreamOfFirst (aY,  size);
    FGetCreamOfFirst (aCb, size);
    FGetCreamOfFirst (aCr, size);
    size := size*SizeOf(Integer);
    Inc (Integer(aY),  size);
    Inc (Integer(aCb), size);
    Inc (Integer(aCr), size);
  end;
  // uncream
  UncreamizeAll;
end;

procedure TPPH.LoadFromFile (filename: string);
var
  port: TPPHbits;
begin
  port := TPPHbits.Create;
  port.LoadFromFile (filename);
  LoadCream (port);
  port.ClearBits;
  port.Destroy;
end;

{
function TPPH.SaveCreamOfSize (size_limit: Integer; dest: TPPH): PIntegerArray;
var
  Y_first, Cb_first, Cr_first: Integer;
begin
  Y_first := FFindFirst (FqY, size_limit*2 div 3);
  Cb_first := FFindFirst (FqCb, size_limit*1 div 6);
  Cr_first := FFindFirst (FqCr, size_limit*1 div 6);
  //
  FS_Q := 0;
  //
  dest.SourceSize := SourceSize;
  FPutBitInit;
  // source size
  FPutAbsValue (FSourceSize.x, $fff);
  FPutAbsValue (FSourceSize.y, $fff);
  // Y
  FPutCream (FqY, dest.qY, Y_first);
  // Cb
  FPutCream (FqCb, dest.qCb, Cb_first);
  // Cr
  FPutCream (FqCr, dest.qCr, Cr_first);
  //
  Result := Fbits_array;
end;
}

end.
