unit pph_viewer_unit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls, pph;

type
  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    miFile: TMenuItem;
    miOpen: TMenuItem;
    miSave: TMenuItem;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    Image1: TImage;
    miHelp: TMenuItem;
    miAbout: TMenuItem;
    Image2: TImage;
    procedure miOpenClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure miSaveClick(Sender: TObject);
    procedure miAboutClick(Sender: TObject);
  private
    { Private declarations }
    FTitle: string;
    procedure FSetTitle (Atitle: string);
    procedure PutBitmap;
    procedure OpenBMP (filename: string);
    procedure OpenPPH (filename: string);
    procedure SaveBMP (filename: string);
    procedure SavePPH (filename: string);
  public
    { Public declarations }
    Fbmp: TBitmap;
    Fpph: TPPH;
    property Title: string read FTitle write FSetTitle;
  end;

var
  Form1: TForm1;

implementation

uses pph_viewer_control;

{$R *.DFM}
procedure TForm1.FSetTitle (Atitle: string);
var
  n: Integer;
begin
  n := Length (Atitle);
  while (n > 0) and (copy(Atitle, n, 1) <> '\') do
    Dec (n);
  Atitle := copy (Atitle, n + 1, Length (Atitle) - n);
  //
  FTitle := Atitle;
  Caption := Atitle + ' - pph viewer';
  Application.Title := Caption;
end;

procedure TForm1.PutBitmap;
begin
  ClientWidth := Fbmp.Width;
  ClientHeight := Fbmp.Height;
  Image1.Canvas.Draw (0,0, Fbmp);
end;

procedure TForm1.OpenBMP (filename: string);
begin
  Fbmp.LoadFromFile (filename);
  Fpph.FromBitmap (Fbmp);
  Title := filename;
  PutBitmap;
end;

procedure TForm1.OpenPPH (filename: string);
begin
  Title := filename;
  {if }Fpph.LoadFromFile (filename);// then begin;
    Fpph.ToBitmap (Fbmp);
    PutBitmap;
  //end;
end;

procedure TForm1.SaveBMP (filename: string);
begin
  Fbmp.SaveToFile (filename);
  Title := filename;
end;

procedure TForm1.SavePPH (filename: string);
var
  n: Integer;
begin
  n := Length (filename);
  while (n > 0) and (copy(filename, n, 1) <> '\') do
    Dec (n);
  filename := copy (filename, n + 1, Length(filename) - n);
  PPHControlForm.Caption := 'phh control - ' + filename;
  PPHControlForm.Show;
end;

// form

procedure TForm1.miOpenClick(Sender: TObject);
var
  ext: string;
begin
  if OpenDialog1.Execute then begin
    ext := copy (OpenDialog1.FileName, Length(OpenDialog1.FileName)-2, 3);
    if ext = 'bmp' then
      OpenBMP (OpenDialog1.FileName)
    else if ext = 'pph' then
      OpenPPH (OpenDialog1.FileName);
  end;
end;

procedure TForm1.miSaveClick(Sender: TObject);
var
  ext: string;
begin
  if Fbmp.Width = 0 then begin
    MessageDlg ('no source :(', mtWarning, [mbOK], 0);
    Exit;
  end;
  if SaveDialog1.Execute then begin
    ext := copy (SaveDialog1.FileName, Length(SaveDialog1.FileName)-2, 3);
    if ext = 'bmp' then
      SaveBMP (SaveDialog1.FileName)
    else if ext = 'pph' then
      SavePPH (SaveDialog1.FileName);
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
{var
  filename: string;}
begin
  Fbmp := TBitmap.Create;
  Fpph := TPPH.Create;
  // if setted param
  if ParamCount = 1 then
    OpenPPH (ParamStr(1))
  else
    miOpenClick (Sender);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  Fbmp.Destroy;      
  Fpph.Destroy;
end;

procedure TForm1.miAboutClick(Sender: TObject);
begin
  MessageDlg ('created by popitch(@mail.ru)', mtWarning, [mbOK], 0);
end;

end.
