unit pphbits;

interface

uses
  Classes, SysUtils, pphtypes;

type
  TPPHbits = class (TObject)
  private
    Fbits_array: PIntegerArray;
    Fbits_size, Fbits_pos: Integer;
    FS_Q: Integer;
    Fbits_1, Fbits_0: Integer;
    procedure FArrayInc;
  public
    constructor Create;
    destructor Destroy; override;
    procedure ClearBits;
    // write
    procedure PutBitInit;
    procedure PutBit (bit: Boolean);
    procedure PutAbsValue (value, maxValue: Integer);
    procedure PutDelta (size,num, size2,num2: Integer);
    procedure CompressBitsR (p: PIntegerArray; pos,size,num: Integer);
    procedure CompressBits (p: PIntegerArray; size, num: Integer);
    // read
    procedure GetBitInit (a: PIntegerArray);
    function GetBit: Boolean;
    function GetAbsValue (maxValue: Integer): Integer;
    function GetDelta (size,num, size2: Integer): Integer;
    procedure DecompressBitsR (p: PIntegerArray; pos,size,num: Integer);
    function DecompressBits (size: Integer): PIntegerArray;
    procedure LoadFromFile (filename: string);
    procedure SaveToFile (filename: string);
    //
    property bits_pos: Integer read Fbits_pos;
    property bits_array: PIntegerArray read Fbits_array;
  end;
var
  Cmn_value: array [1..16] of PIntegerArray;//1    2    3     4    5    6    7   8   9   10  11  12  13  14  15  16
  Cmn_limit: array [1..16] of Integer       =(2345,2345,2345, 477, 193, 110, 75, 58, 49, 43, 39, 37, 35, 34, 34, 33);

implementation

{ bits }

constructor TPPHbits.Create;
begin
  inherited Create;
  Fbits_array := nil;
end;

destructor TPPHbits.Destroy;
begin
  //ClearBits;
  inherited Destroy;
end;

procedure TPPHbits.ClearBits;
begin
  if Fbits_array <> nil then begin
    FreeMem (Fbits_array);
    Fbits_array := nil;
  end;
end;

// write

procedure TPPHbits.PutBitInit;
var
  n: Integer;
begin
  // clear
  ClearBits;
  // init
    Fbits_array := AllocMem ($100*SizeOf(Integer));
    for n := 0 to $ff do
      Fbits_array[n] := 0;
    Fbits_size := 32*$100;
    Fbits_pos := 0;
  //
  Fbits_0 := 0;
  Fbits_1 := 0;
end;

procedure TPPHbits.FArrayInc;
var
  n: Integer;
begin
  Inc (Fbits_size, 32*$100);
  ReAllocMem (Fbits_array, (Fbits_size shr 5)*SizeOf(Integer));
  for n := (Fbits_size shr 5)-$100 to (Fbits_size shr 5)-1 do
    Fbits_array[n] := 0;
end;

procedure TPPHbits.PutBit (bit: Boolean);
var
  n: Integer;
begin
  // put bit
  if Fbits_pos = Fbits_size then
    FArrayInc;
  if bit then begin
    Inc (Fbits_array[Fbits_pos shr 5], 1 shl (Fbits_pos and $1f));
    Inc (Fbits_1);
  end else
    Inc (Fbits_0);
  Inc (Fbits_pos);
end;

procedure TPPHbits.PutAbsValue (value, maxValue: Integer);
var
  n: Integer;
begin
  n := 0;
  while maxValue <> 0 do begin
    PutBit ((value and (1 shl n)) <> 0);
    maxValue := maxValue and (-(1 shl n)-1);
    Inc (n);
  end;
end;

procedure TPPHbits.PutDelta (size,num, size2,num2: Integer);
var
  num2_min, num2_max: Integer;
begin
  // find min value of num2
  num2_min := num - (size-size2);
  if num2_min < 0 then
    num2_min := 0;
  // find max value of num2
  if num < size2 then
    num2_max := num
  else
    num2_max := size2;
  // put num2
  PutAbsValue (num2 - num2_min, num2_max - num2_min);
end;

procedure TPPHbits.CompressBitsR (p: PIntegerArray; pos,size,num: Integer);
var
  toFind: Boolean;
  vnum, n: Integer;
  v, vmax: Integer;
  num2,size2: Integer;
begin
  if (num=0) or (num=size) then
    Exit;
  // what for find?
  if num < (size-num) then begin
    toFind := True;
    vnum := num;
  end else begin
    toFind := False;
    vnum := size-num;
  end;
  //
  if vnum = 1 then begin
    // one bit = toFind
    n := pos;
    while ((p[n shr 5] and (1 shl (n and $1f))) <> 0) xor toFind do
      Inc (n);
    v := n - pos;
    vmax := size-1;
    PutAbsValue (v, vmax);
  end else if (vnum = 2) and (size <= 46340) then begin
    // two bits = toFind
    v := 0;
    n := pos;
    while ((p[n shr 5] and (1 shl (n and $1f))) <> 0) xor toFind do begin
      Inc (n);
      Inc (v, size+pos - n);
    end;
    Inc (n);
    while ((p[n shr 5] and (1 shl (n and $1f))) <> 0) xor toFind do begin
      Inc (n);
      Inc (v);
    end;
    vmax := (size*(size-1) div 2) - 1;
    PutAbsValue (v, vmax);
  end else if (vnum <= 16) and (size <= Cmn_limit[vnum]) then begin
    // >2 bits = toFind..
    vmax := Cmn_value[vnum][size];
    v := 0;
    while vnum <> 0 do begin
      n := pos;
      while ((p[n shr 5] and (1 shl (n and $1f))) <> 0) xor toFind do
        Inc (n);
      n := n-pos + 1;
      Dec (size, n);
      Inc (v, Cmn_value[vnum][size]);
      Dec (vnum);
      Inc (pos, n);
    end;
    PutAbsValue (v, vmax);
  end else if size > 20 then begin
    // divide
    size2 := size div 2;
    num2 := 0;
    for n := pos to pos+size2-1 do
      if (p[n shr 5] and (1 shl (n and $1f))) <> 0 then
        Inc (num2);
    //FPutAbsValue (num2, num);
    PutDelta (size,num, size2,num2);
    CompressBitsR (p, pos, size2, num2);
    CompressBitsR (p, pos+size2, size-size2, num-num2);
  end else begin
    // put all bits
    for n := pos to pos+size-1 do
      if (p[n shr 5] and (1 shl (n and $1f))) <> 0 then
        PutBit (True)
      else
        PutBit (False);
  end;
end;

procedure TPPHbits.CompressBits (p: PIntegerArray; size, num: Integer);
begin
  PutAbsValue (num, size);
  CompressBitsR (p, 0, size, num);
end;

// bits read

procedure TPPHbits.GetBitInit (a: PIntegerArray);
begin
  // clear
  ClearBits;
  // get bit init
  Fbits_array := a;
  Fbits_pos := 0;
  //
  Fbits_0 := 0;
  Fbits_1 := 0;
end;

function TPPHbits.GetBit: Boolean;
begin
  // get bit
  if (Fbits_array[Fbits_pos shr 5] and (1 shl (Fbits_pos and $1f))) <> 0 then begin
    Result := True;
    Inc (Fbits_1);
  end else begin
    Result := False;
    Inc (Fbits_0);
  end;
  Inc (Fbits_pos);
end;

function TPPHbits.GetAbsValue (maxValue: Integer): Integer;
var
  n: Integer;
begin
  n := 0;
  Result := 0;
  while maxValue <> 0 do begin
    if GetBit then
      Inc (Result, 1 shl n);
    maxValue := maxValue shr 1;
    Inc (n);
  end;
end;

function TPPHbits.GetDelta (size,num, size2: Integer): Integer;
var
  num2_min, num2_max: Integer;
begin
  // find min value of num2
  num2_min := num - (size-size2);
  if num2_min < 0 then
    num2_min := 0;
  // find max value of num2
  if num < size2 then
    num2_max := num
  else
    num2_max := size2;
  // put num2
  Result := num2_min + GetAbsValue (num2_max - num2_min);
end;

procedure TPPHbits.DecompressBitsR (p: PIntegerArray; pos,size,num: Integer);
var
  toFind: Boolean;
  vnum, n: Integer;
  v, vmax: Integer;
  num2,size2: Integer;
begin
  if size = 0 then
    Exit;
  {else if num=size then
    for n := pos to pos+size-1 do
      Inc (p[n shr 5], 1 shl (n and $1f));}
  // what for find?
  if num < (size-num) then begin
    toFind := True;
    vnum := num;
  end else begin
    toFind := False;
    vnum := size-num;
  end;
  // all bits ==
  if vnum = 0 then begin
    if not toFind then
      for n := pos to pos+size-1 do
        Inc (p[n shr 5], 1 shl (n and $1f));
    Exit;
  end;
  //
  if vnum = 1 then begin
    // one bit = toFind
    vmax := size-1;
    v := pos + GetAbsValue (vmax);
    if toFind then
      Inc (p[v shr 5], 1 shl (v and $1f))
    else
      for n := pos to pos+size-1 do
        if n <> v then
          Inc (p[n shr 5], 1 shl (n and $1f));
  end else if (vnum = 2) and (size <= 46340) then begin
    // two bits = toFind
    vmax := (size*(size-1) div 2) - 1;
    v := GetAbsValue (vmax);
    n := pos;
    // first
    while v >= (size+pos-n-1) do begin
      {if not toFind then
        Inc (p[n shr 5], 1 shl (n and $1f));}
      Inc (n);
      Dec (v, size+pos-n{?-1});
    end;
    if toFind then begin
      Inc (p[n shr 5], 1 shl (n and $1f));
      // second
      Inc (n);
      Inc (n, v);
      Inc (p[n shr 5], 1 shl (n and $1f));
    end else begin
      num2 := n;
      for n := pos to pos+size-1 do
        if (n <> num2) and (n <> (num2+1+v)) then
          Inc (p[n shr 5], 1 shl (n and $1f));
    end;
  end else if (vnum <= 16) and (size <= Cmn_limit[vnum]) then begin
    // >2 bits = toFind..
    vmax := Cmn_value[vnum][size];
    v := GetAbsValue (vmax)+1;
    vmax := pos+size;
    n := pos;
    while vnum <> 0 do begin
      Dec (size);
      while Cmn_value[vnum][size] >={?} v do begin
        if not toFind then
          Inc (p[n shr 5], 1 shl (n and $1f));
        Inc (n);
        Dec (size);
      end;
      {?}Dec (v, Cmn_value[vnum][size]);{}
      if toFind then
        Inc (p[n shr 5], 1 shl (n and $1f));
      Inc (n);
      Dec (vnum);
    end;
    if not toFind then
      while n < vmax do begin
        Inc (p[n shr 5], 1 shl (n and $1f));
        Inc (n);
      end;
  end else if size > 20 then begin
    // divide
    size2 := size div 2;
    //num2 := FGetAbsValue (num);
    num2 := GetDelta (size,num, size2);
    DecompressBitsR (p, pos, size2, num2);
    DecompressBitsR (p, pos+size2, size-size2, num-num2);
  end else begin
    // get all bits
    for n := pos to pos+size-1 do
      if GetBit then
        Inc (p[n shr 5], 1 shl (n and $1f));
  end;
end;

function TPPHbits.DecompressBits (size: Integer): PIntegerArray;
var
  num: Integer;
begin
  Result := AllocIntArray ((size+31) div 32);
  for num := 0 to ((size+31) div 32)-1 do
    Result[num] := 0;
  num := GetAbsValue (size);
  DecompressBitsR (Result, 0, size, num);
end;

procedure TPPHbits.LoadFromFile (filename: string);
var
  a: PIntegerArray;
  f: TFileStream;
  isize, n: Integer;
begin
  f := TFileStream.Create (filename, fmOpenRead);
  f.Read (isize, SizeOf(Integer));
  a := AllocIntArray (isize);
  for n := 0 to isize-1 do
    f.Read (a[n], SizeOf(Integer));
  f.Destroy;
  Fbits_array := a;
  Fbits_pos := 0;
end;

procedure TPPHbits.SaveToFile (filename: string);
var
  f: TFileStream;
  isize, n: Integer;
begin
  isize := (Fbits_size + 31) div 32;
  f := TFileStream.Create (filename, fmCreate);
  f.Write (isize, SizeOf(Integer));
  for n := 0 to isize-1 do
    f.Write (Fbits_array[n], SizeOf(Integer));
  f.Destroy;
end;

{ Cmn for bits }

procedure CmnInitialization;
var
  m, n: Integer;
begin
  Cmn_value[1] := AllocMem ((Cmn_limit[1]+1)*SizeOf(Integer));
  for n := 0 to Cmn_limit[1] do
    Cmn_value[1][n] := n;
  for m := 2 to 16 do begin
    Cmn_value[m] := AllocMem ((Cmn_limit[m]+1)*SizeOf(Integer));
    Cmn_value[m][m-1] := 0;
    for n := m to Cmn_limit[m] do
       Cmn_value[m][n] := Cmn_value[m][n-1] + Cmn_value[m-1][n-1];
  end;
end;

procedure CmnFinalization;
var
  m: Integer;
begin
  for m := 1 to 16 do
    FreeMem (Cmn_value[m]);
end;

initialization
  CmnInitialization;

finalization
  CmnFinalization;

end.
