unit pphtypes;

interface

uses
  SysUtils;

type
  TRGB = record
    R, G, B: Byte;
  end;
  TRGBArray = array [0..$ffffff] of TRGB;
  PRGBArray = ^TRGBArray;

  TYCbCr = record
    Y, Cb, Cr: Integer;
  end;
  TYCbCrArray = array [0..$ffffff] of TYCbCr;
  PYCbCrArray = ^TYCbCrArray;

  TByteArray = array [0..$ffffff] of Byte;
  PByteArray = ^TByteArray;
  TIntegerArray = array [0..$ffffff] of Integer;
  PIntegerArray = ^TIntegerArray;
  TIntegerArrays = array [0..$ffffff] of PIntegerArray;
  PIntegerArrays = ^TIntegerArrays;

function AllocIntArray (size: Integer): PIntegerArray;
function AllocIntArrays (width,height: Integer): PIntegerArrays;
function YCbCr (AY,ACb,ACr: Byte): TYCbCr;
function RGB2YCbCr (source: TRGB): TYCbCr;
function YCbCr2RGB (source: TYCbCr): TRGB;

implementation

function AllocIntArray (size: Integer): PIntegerArray;
begin
  Result := AllocMem (size*SizeOf(Integer));
end;

function AllocIntArrays (width,height: Integer): PIntegerArrays;
var
  n: Integer;
begin
  Result := AllocMem (height*(width*SizeOf(Integer) + SizeOf(PIntegerArray)));
  for n := 0 to height-1 do
    Integer (Result[n]) := Integer (Result) + height*SizeOf(PIntegerArray) +
      n*width*SizeOf(Integer);
end;

function YCbCr (AY,ACb,ACr: Byte): TYCbCr;
begin
  Result.Y := AY;
  Result.Cb := ACb;
  Result.Cr := ACr;
end;

function RGB2YCbCr (source: TRGB): TYCbCr;
begin
  Result.Y  := Trunc (  0.299*source.R +  0.587*source.G + 0.114*source.B);
  Result.Cb := Trunc (-0.1687*source.R - 0.3313*source.G + 0.5*source.B);
  Result.Cr := Trunc (    0.5*source.R - 0.4187*source.G - 0.0813*source.B);
end;

function YCbCr2RGB (source: TYCbCr): TRGB;
var
  i: Integer;
begin
  i := Trunc (source.Y + 1.402*source.Cr);
  if i < 0 then i := 0;
  if i > 255 then i := 255;
  Result.R := i;
  //
  i := Trunc (source.Y - 0.34414*source.Cb - 0.71414*source.Cr);
  if i < 0 then i := 0;
  if i > 255 then i := 255;
  Result.G := i;
  //
  i := Trunc (source.Y + 1.772*source.Cb);
  if i < 0 then i := 0;
  if i > 255 then i := 255;
  Result.B := i;
end;

end.
